# encoding: utf-8
"""
created: August, 2019

author: P. R. Wiecha

Part of set of example scripts demonstrating `pyGDM2.EO` for directional 
scattering problems.

script "04": evoltionary optimization of dipole emitter directional antenna. 
             Target: direct emission of X oriented dipole towards OX.
             Constrain optimization to force a mirror symmetry around X-axis.
"""
import numpy as np

from pyGDM2 import structures
from pyGDM2 import materials
from pyGDM2 import fields
from pyGDM2 import tools
from pyGDM2 import core
import pyGDM2.EO.models as eo_models
import pyGDM2.EO.problems as eo_problems
from pyGDM2.EO.core import run_eo




# =============================================================================
# directionality config -- emission towards OX
# =============================================================================
Nteta = 5
Nphi=12
tetamin=0
tetamax=np.pi/2.
dir_index = [48]

#==============================================================================
# Setup pyGDM part
#==============================================================================
## ---------- Setup structure 
## (use very coarse mesh for fastest evaluation, gives similar results)
mesh = 'cube'
step = 40
material = materials.gold()
n1, n2 = 1.0, 1.0         # in vacuum

## --- empty dummy-geometry, will be set during run-time by EO
geometry = []       
struct = structures.struct(step, geometry, material, n1,n2, structures.get_normalization(mesh))


## ---------- setup incident field
field_generator = fields.dipole_electric
wavelengths = [800]
kwargs = dict(x0=0, y0=0, z0=20.0, 
              mx=1, my=0, mz=0)
efield = fields.efield(field_generator, wavelengths=wavelengths, kwargs=kwargs)



## ---------- simulation initialization
sim = core.simulation(struct, efield)




#==============================================================================
# setup evolutionary optimization
#==============================================================================
## --- save file-name
file_out = "eo_direct_gold_Xdipole_symmetric.eo"

## --- structure model and optimizaton problem
model = eo_models.BlockModel(sim, 
                             N=20, # total nr of blocks (if symmetric: x2)
                             block_nx=1, block_ny=1, block_nz=1, # block size in diples
                             area_limits=[-12,12],  # 1 micron^2 large allowed area
                             forbidden=[[0,0]],  # forbidden positions (here: dipole x,y location)
                             symmetric=True)

## --- problem: scattered intensity towards phi=0 deg / teta=90 deg
problem = eo_problems.ProblemDirectivity(model,
                 dir_index=dir_index, 
                 kwargs_farfield=dict(Nteta=Nteta, Nphi=Nphi, 
                                      tetamin=tetamin, tetamax=tetamax),
                 which_field='e_tot', # total field: dipole emission + scattering
                 kwargs_scatter=dict(method='lu'), # kwargs passed to "core.scatter
                 consider_dS=False,  # true integral or unweighted sum? 'False' better for coarse discretization.
                 averaging=True,     # average intensity per window?
                 absoluteI=False)    # maximize absolute scattered I_dir or ratio I_dir/I_rest ?


## --- size of population
population = 20          # Small for testing, but high risk to get stuck in local extrema.


## --- stop criteria
max_time = 1*3600         # in seconds
max_iter = 200            # max. iterations
max_nonsuccess = 100      # max. consecutive iterations without improvement


## --- other config
generations = 1          # generations to evolve between status reports
plot_interval = 1        # plot each N improvements
save_all_generations = False



##  Use algorithm "sade" (jDE variant, a self-adaptive form of differential evolution)
import pygmo as pg
algorithm = pg.sade
algorithm_kwargs = dict()   # empty: use default config of algorithm

eo_dict = run_eo(problem,
                 population=population,
                 algorithm=algorithm, algorithm_kwargs=algorithm_kwargs,
                 plot_interval=plot_interval, 
                 generations=generations, save_all_generations=save_all_generations,
                 max_time=max_time, max_iter=max_iter, max_nonsuccess=max_nonsuccess,
                 filename=file_out)



